import re
from typing import TYPE_CHECKING, Any, cast

from litestar import MediaType
from litestar.connection import Request
from litestar.connection.base import AuthT, StateT, UserT
from litestar.exceptions import (
    HTTPException,
    ImproperlyConfiguredException,
    InternalServerException,
    NotAuthorizedException,
    NotFoundException,
    PermissionDeniedException,
)
from litestar.exceptions.responses import (
    create_debug_response,  # pyright: ignore[reportUnknownVariableType]
    create_exception_response,  # pyright: ignore[reportUnknownVariableType]
)
from litestar.plugins.flash import flash
from litestar.repository.exceptions import (
    ConflictError,  # pyright: ignore[reportUnknownVariableType,reportAttributeAccessIssue]
    NotFoundError,  # pyright: ignore[reportUnknownVariableType,reportAttributeAccessIssue]
    RepositoryError,  # pyright: ignore[reportUnknownVariableType,reportAttributeAccessIssue]
)
from litestar.response import Response
from litestar.status_codes import (
    HTTP_400_BAD_REQUEST,
    HTTP_401_UNAUTHORIZED,
    HTTP_409_CONFLICT,
    HTTP_422_UNPROCESSABLE_ENTITY,
    HTTP_500_INTERNAL_SERVER_ERROR,
)

from litestar_vite.inertia.response import InertiaBack, InertiaRedirect, InertiaResponse, error

if TYPE_CHECKING:
    from litestar_vite.inertia.plugin import InertiaPlugin

FIELD_ERR_RE = re.compile(r"field `(.+)`$")


class _HTTPConflictException(HTTPException):
    """Request conflict with the current state of the target resource."""

    status_code: int = HTTP_409_CONFLICT


def exception_to_http_response(request: Request[UserT, AuthT, StateT], exc: Exception) -> Response[Any]:
    """Handler for all exceptions subclassed from HTTPException."""
    inertia_enabled = getattr(request, "inertia_enabled", False) or getattr(request, "is_inertia", False)

    if not inertia_enabled:
        if isinstance(exc, NotFoundError):
            http_exc = NotFoundException
        elif isinstance(exc, (RepositoryError, ConflictError)):
            http_exc = _HTTPConflictException  # type: ignore[assignment]
        else:
            http_exc = InternalServerException  # type: ignore[assignment]
        if request.app.debug and http_exc not in (PermissionDeniedException, NotFoundError):
            return cast("Response[Any]", create_debug_response(request, exc))
        return cast("Response[Any]", create_exception_response(request, http_exc(detail=str(exc.__cause__))))  # pyright: ignore[reportUnknownArgumentType]
    return create_inertia_exception_response(request, exc)


def create_inertia_exception_response(request: Request[UserT, AuthT, StateT], exc: Exception) -> Response[Any]:
    """Create the inertia exception response"""
    is_inertia = getattr(request, "is_inertia", False)
    status_code = getattr(exc, "status_code", HTTP_500_INTERNAL_SERVER_ERROR)
    preferred_type = MediaType.HTML if not is_inertia else MediaType.JSON
    detail = getattr(exc, "detail", "")  # litestar exceptions
    extras = getattr(exc, "extra", "")  # msgspec exceptions
    content = {"status_code": status_code, "message": getattr(exc, "detail", "")}
    inertia_plugin = cast("InertiaPlugin", request.app.plugins.get("InertiaPlugin"))
    if extras:
        content.update({"extra": extras})
    try:
        flash(request, detail, category="error")
    except (AttributeError, ImproperlyConfiguredException):
        msg = "Unable to set `flash` session state.  A valid session was not found for this request."
        request.logger.warning(msg)
    if extras and len(extras) >= 1:
        message = extras[0]
        default_field = f"root.{message.get('key')}" if message.get("key", None) is not None else "root"  # type: ignore
        error_detail = cast("str", message.get("message", detail))  # type: ignore[union-attr] # pyright: ignore[reportUnknownMemberType]
        match = FIELD_ERR_RE.search(error_detail)
        field = match.group(1) if match else default_field
        if isinstance(message, dict):
            error(request, field, error_detail if error_detail else detail)
    if status_code in {HTTP_422_UNPROCESSABLE_ENTITY, HTTP_400_BAD_REQUEST}:
        return InertiaBack(request)
    if isinstance(exc, PermissionDeniedException):
        return InertiaBack(request)
    if (status_code == HTTP_401_UNAUTHORIZED or isinstance(exc, NotAuthorizedException)) and (
        inertia_plugin.config.redirect_unauthorized_to is not None
        and str(request.url) != inertia_plugin.config.redirect_unauthorized_to
    ):
        return InertiaRedirect(request, redirect_to=inertia_plugin.config.redirect_unauthorized_to)
    return InertiaResponse[Any](
        media_type=preferred_type,
        content=content,
        status_code=status_code,
    )



from pathlib import Path

import pytest

from litestar_vite.config import ViteConfig
from litestar_vite.inertia.config import InertiaConfig
from litestar_vite.inertia.plugin import InertiaPlugin
from litestar_vite.plugin import VitePlugin

pytestmark = pytest.mark.anyio


@pytest.fixture
def inertia_config() -> InertiaConfig:
    return InertiaConfig(root_template="index.html.j2")


@pytest.fixture
def inertia_plugin(inertia_config: InertiaConfig) -> InertiaPlugin:
    return InertiaPlugin(config=inertia_config)


@pytest.fixture
def vite_plugin() -> VitePlugin:
    return VitePlugin(
        config=ViteConfig(
            resource_dir=Path(__file__).parent / "resources",
            template_dir=Path(__file__).parent / "templates",
        ),
    )



from __future__ import annotations

import os
from contextlib import contextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Iterator, cast

from litestar.plugins import CLIPlugin, InitPluginProtocol
from litestar.static_files import (
    create_static_files_router,  # pyright: ignore[reportUnknownVariableType]
)

from litestar_vite.config import ViteConfig

if TYPE_CHECKING:
    from click import Group
    from litestar import Litestar
    from litestar.config.app import AppConfig

    from litestar_vite.config import ViteTemplateConfig
    from litestar_vite.template_engine import ViteTemplateEngine


def set_environment(config: ViteConfig) -> None:
    """Configure environment for easier integration"""
    os.environ.setdefault("ASSET_URL", config.asset_url)
    os.environ.setdefault("VITE_ALLOW_REMOTE", str(True))
    os.environ.setdefault("VITE_PORT", str(config.port))
    os.environ.setdefault("VITE_HOST", config.host)
    os.environ.setdefault("VITE_PROTOCOL", config.protocol)
    os.environ.setdefault("APP_URL", f"http://localhost:{os.environ.get('LITESTAR_PORT',8000)}")
    if config.dev_mode:
        os.environ.setdefault("VITE_DEV_MODE", str(config.dev_mode))


class VitePlugin(InitPluginProtocol, CLIPlugin):
    """Vite plugin."""

    __slots__ = ("_config",)

    def __init__(self, config: ViteConfig | None = None) -> None:
        """Initialize ``Vite``.

        Args:
            config: configuration to use for starting Vite.  The default configuration will be used if it is not provided.
        """
        if config is None:
            config = ViteConfig()
        self._config = config

    @property
    def config(self) -> ViteConfig:
        return self._config

    @property
    def template_config(self) -> ViteTemplateConfig[ViteTemplateEngine]:
        from litestar_vite.config import ViteTemplateConfig
        from litestar_vite.template_engine import ViteTemplateEngine

        return ViteTemplateConfig[ViteTemplateEngine](
            engine=ViteTemplateEngine,
            config=self._config,
            directory=self._config.template_dir,
        )

    def on_cli_init(self, cli: Group) -> None:
        from litestar_vite.cli import vite_group

        cli.add_command(vite_group)

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Configure application for use with Vite.

        Args:
            app_config: The :class:`AppConfig <.config.app.AppConfig>` instance.
        """

        if self._config.template_dir is not None:
            app_config.template_config = self.template_config

        if self._config.set_static_folders:
            static_dirs = [Path(self._config.bundle_dir), Path(self._config.resource_dir)]
            if Path(self._config.public_dir).exists() and self._config.public_dir != self._config.bundle_dir:
                static_dirs.append(Path(self._config.public_dir))
            app_config.route_handlers.append(
                create_static_files_router(
                    directories=cast(  # type: ignore[arg-type]
                        "list[Path]",
                        static_dirs if self._config.dev_mode else [Path(self._config.bundle_dir)],
                    ),
                    path=self._config.asset_url,
                    name="vite",
                    html_mode=False,
                    include_in_schema=False,
                    opt={"exclude_from_auth": True},
                ),
            )
        return app_config

    @contextmanager
    def server_lifespan(self, app: Litestar) -> Iterator[None]:
        import threading

        from litestar.cli._utils import console

        from litestar_vite.commands import execute_command

        if self._config.use_server_lifespan and self._config.dev_mode:
            command_to_run = self._config.run_command if self._config.hot_reload else self._config.build_watch_command
            if self.config.hot_reload:
                console.rule("[yellow]Starting Vite process with HMR Enabled[/]", align="left")
            else:
                console.rule("[yellow]Starting Vite watch and build process[/]", align="left")
            if self._config.set_environment:
                set_environment(config=self._config)
            vite_thread = threading.Thread(
                name="vite",
                target=execute_command,
                args=[],
                kwargs={"command_to_run": command_to_run, "cwd": self._config.root_dir},
            )
            try:
                vite_thread.start()
                yield
            finally:
                if vite_thread.is_alive():
                    vite_thread.join(timeout=5)
                console.print("[yellow]Vite process stopped.[/]")

        else:
            manifest_path = Path(f"{self._config.bundle_dir}/{self._config.manifest_name}")
            if manifest_path.exists():
                console.rule(f"[yellow]Serving assets using manifest at `{manifest_path!s}`.[/]", align="left")
            else:
                console.rule(f"[yellow]Serving assets without manifest at `{manifest_path!s}`.[/]", align="left")
            yield



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Generic, TypedDict, TypeVar

__all__ = (
    "InertiaHeaderType",
    "PageProps",
)


T = TypeVar("T")


@dataclass
class PageProps(Generic[T]):
    """Inertia Page Props Type."""

    component: str
    url: str
    version: str
    props: dict[str, Any]


@dataclass
class InertiaProps(Generic[T]):
    """Inertia Props Type."""

    page: PageProps[T]


class InertiaHeaderType(TypedDict, total=False):
    """Type for inertia_headers parameter in get_headers()."""

    enabled: bool | None
    version: str | None
    location: str | None
    partial_data: str | None
    partial_component: str | None



from __future__ import annotations

APP_DEFAULT_CONFIG_FILE_CONTENT = """
from litestar import Litestar

from litestar_vite import VitePlugin

app = Litestar(plugins=[VitePlugin()])
"""


APP_BASIC_NO_ROUTES_FILE_CONTENT = """
from litestar import Litestar

from litestar_vite import ViteConfig, VitePlugin

vite = VitePlugin(config=ViteConfig(template_dir="{template_dir!s}"))

app = Litestar(plugins=[vite])
"""



from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from typing import TYPE_CHECKING

from litestar.app import DEFAULT_OPENAPI_CONFIG
from litestar.cli._utils import (
    remove_default_schema_routes,
    remove_routes_with_patterns,
)
from litestar.routes import ASGIRoute, WebSocketRoute
from litestar.serialization import encode_json

if TYPE_CHECKING:
    from litestar import Litestar


@dataclass(frozen=True)
class Routes:
    routes: dict[str, str]

    @cached_property
    def formatted_routes(self) -> str:
        return encode_json(self.routes).decode(encoding="utf-8")


EXCLUDED_METHODS = {"HEAD", "OPTIONS", "TRACE"}


def generate_js_routes(
    app: Litestar,
    exclude: tuple[str, ...] | None = None,
    schema: bool = False,
) -> Routes:
    sorted_routes = sorted(app.routes, key=lambda r: r.path)
    if not schema:
        openapi_config = app.openapi_config or DEFAULT_OPENAPI_CONFIG
        sorted_routes = remove_default_schema_routes(sorted_routes, openapi_config)
    if exclude is not None:
        sorted_routes = remove_routes_with_patterns(sorted_routes, exclude)
    route_list: dict[str, str] = {}
    for route in sorted_routes:
        if isinstance(route, (ASGIRoute, WebSocketRoute)):
            route_name = route.route_handler.name or route.route_handler.handler_name
            if len(route.methods.difference(EXCLUDED_METHODS)) > 0:
                route_list[route_name] = route.path
        else:
            for handler in route.route_handlers:
                route_name = handler.name or handler.handler_name
                if handler.http_methods.isdisjoint(EXCLUDED_METHODS):
                    route_list[route_name] = route.path

    return Routes(routes=route_list)



from __future__ import annotations

from typing import TYPE_CHECKING

from litestar.exceptions import ImproperlyConfiguredException
from litestar.middleware import DefineMiddleware
from litestar.middleware.session import SessionMiddleware
from litestar.plugins import InitPluginProtocol
from litestar.security.session_auth.middleware import MiddlewareWrapper
from litestar.utils.predicates import is_class_and_subclass

from litestar_vite.inertia.exception_handler import exception_to_http_response
from litestar_vite.inertia.middleware import InertiaMiddleware
from litestar_vite.inertia.request import InertiaRequest
from litestar_vite.inertia.response import InertiaResponse
from litestar_vite.inertia.routes import generate_js_routes

if TYPE_CHECKING:
    from litestar import Litestar
    from litestar.config.app import AppConfig

    from litestar_vite.inertia.config import InertiaConfig


def set_js_routes(app: Litestar) -> None:
    """Generate the route structure of the application on startup."""
    js_routes = generate_js_routes(app)
    app.state.js_routes = js_routes


class InertiaPlugin(InitPluginProtocol):
    """Inertia plugin."""

    __slots__ = ("config",)

    def __init__(self, config: InertiaConfig) -> None:
        """Initialize ``Inertia``.

        Args:
            config: configure and start Vite.
        """
        self.config = config

    def on_app_init(self, app_config: AppConfig) -> AppConfig:
        """Configure application for use with Vite.

        Args:
            app_config: The :class:`AppConfig <.config.app.AppConfig>` instance.
        """
        for mw in app_config.middleware:
            if isinstance(mw, DefineMiddleware) and is_class_and_subclass(
                mw.middleware,
                (MiddlewareWrapper, SessionMiddleware),
            ):
                break
        else:
            msg = "The Inertia plugin require a session middleware."
            raise ImproperlyConfiguredException(msg)
        app_config.exception_handlers.update({Exception: exception_to_http_response})  # pyright: ignore[reportUnknownMemberType]
        app_config.request_class = InertiaRequest
        app_config.response_class = InertiaResponse
        app_config.middleware.append(InertiaMiddleware)
        app_config.on_startup.append(set_js_routes)
        return app_config



from .config import InertiaConfig
from .exception_handler import create_inertia_exception_response, exception_to_http_response
from .middleware import InertiaMiddleware
from .plugin import InertiaPlugin
from .request import InertiaDetails, InertiaHeaders, InertiaRequest
from .response import (
    InertiaBack,
    InertiaExternalRedirect,
    InertiaRedirect,
    InertiaResponse,
    error,
    get_shared_props,
    share,
)
from .routes import generate_js_routes

__all__ = (
    "InertiaBack",
    "InertiaConfig",
    "InertiaDetails",
    "InertiaExternalRedirect",
    "InertiaHeaders",
    "InertiaMiddleware",
    "InertiaPlugin",
    "InertiaRedirect",
    "InertiaRequest",
    "InertiaResponse",
    "create_inertia_exception_response",
    "error",
    "exception_to_http_response",
    "generate_js_routes",
    "get_shared_props",
    "share",
)



from __future__ import annotations

import platform
import subprocess
from pathlib import Path
from typing import TYPE_CHECKING, Any, MutableMapping

if TYPE_CHECKING:
    from jinja2 import Environment, Template
    from litestar import Litestar

VITE_INIT_TEMPLATES: set[str] = {"package.json.j2", "tsconfig.json.j2", "vite.config.ts.j2"}
DEFAULT_RESOURCES: set[str] = {"styles.css.j2", "main.ts.j2"}
DEFAULT_DEV_DEPENDENCIES: dict[str, str] = {
    "typescript": "^5.3.3",
    "vite": "^5.3.3",
    "litestar-vite-plugin": "^0.6.2",
    "@types/node": "^20.14.10",
}
DEFAULT_DEPENDENCIES: dict[str, str] = {"axios": "^1.7.2"}


def to_json(value: Any) -> str:
    """Serialize JSON field values.

    Args:
        value: Any json serializable value.

    Returns:
        JSON string.
    """
    from litestar.serialization import encode_json

    return encode_json(value).decode("utf-8")


def init_vite(
    app: Litestar,
    root_path: Path,
    resource_path: Path,
    asset_url: str,
    public_path: Path,
    bundle_path: Path,
    enable_ssr: bool,
    vite_port: int,
    hot_file: Path,
    litestar_port: int,
) -> None:
    """Initialize a new Vite project.

    Args:
        app: The Litestar application instance.
        root_path: Root directory for the Vite project.
        resource_path: Directory containing source files.
        asset_url: Base URL for serving assets.
        public_path: Directory for static files.
        bundle_path: Output directory for built files.
        enable_ssr: Enable server-side rendering.
        vite_port: Port for Vite dev server.
        hot_file: Path to hot reload manifest.
        litestar_port: Port for Litestar server.
    """
    from jinja2 import Environment, FileSystemLoader, select_autoescape
    from litestar.cli._utils import console
    from litestar.utils import module_loader

    template_path = module_loader.module_to_os_path("litestar_vite.templates")
    vite_template_env = Environment(
        loader=FileSystemLoader([template_path]),
        autoescape=select_autoescape(),
    )

    enabled_templates: set[str] = VITE_INIT_TEMPLATES
    enabled_resources: set[str] = DEFAULT_RESOURCES
    dependencies: dict[str, str] = DEFAULT_DEPENDENCIES
    dev_dependencies: dict[str, str] = DEFAULT_DEV_DEPENDENCIES
    templates: dict[str, Template] = {
        template_name: get_template(environment=vite_template_env, name=template_name)
        for template_name in enabled_templates
    }
    for template_name, template in templates.items():
        target_file_name = template_name[:-3] if template_name.endswith(".j2") else template_name
        with Path(target_file_name).open(mode="w") as file:
            console.print(f" * Writing {target_file_name} to {Path(target_file_name)!s}")

            file.write(
                template.render(
                    entry_point=[
                        f"{resource_path!s}/{resource_name[:-3] if resource_name.endswith('.j2') else resource_name}"
                        for resource_name in enabled_resources
                    ],
                    enable_ssr=enable_ssr,
                    asset_url=asset_url,
                    root_path=root_path,
                    resource_path=str(resource_path.relative_to(root_path)),
                    public_path=str(public_path.relative_to(root_path)),
                    bundle_path=str(bundle_path.relative_to(root_path)),
                    hot_file=str(hot_file.relative_to(root_path)),
                    vite_port=str(vite_port),
                    litestar_port=litestar_port,
                    dependencies=to_json(dependencies),
                    dev_dependencies=to_json(dev_dependencies),
                ),
            )

    for resource_name in enabled_resources:
        template = get_template(environment=vite_template_env, name=resource_name)
        target_file_name = f"{resource_path}/{resource_name[:-3] if resource_name.endswith('.j2') else resource_name}"
        with Path(target_file_name).open(mode="w") as file:
            console.print(
                f" * Writing {resource_name[:-3] if resource_name.endswith('.j2') else resource_name} to {Path(target_file_name)!s}",
            )
            file.write(template.render())
    console.print("[yellow]Vite initialization completed.[/]")


def get_template(
    environment: Environment,
    name: str | Template,
    parent: str | None = None,
    globals: MutableMapping[str, Any] | None = None,  # noqa: A002
) -> Template:
    return environment.get_template(name=name, parent=parent, globals=globals)


def execute_command(command_to_run: list[str], cwd: str | Path | None = None) -> subprocess.CompletedProcess[bytes]:
    """Run Vite in a subprocess."""
    kwargs = {}
    if cwd is not None:
        kwargs["cwd"] = Path(cwd)
    return subprocess.run(command_to_run, check=False, shell=platform.system() == "Windows", **kwargs)  # type: ignore[call-overload]



from __future__ import annotations

import itertools
from collections import defaultdict
from functools import lru_cache
from mimetypes import guess_type
from pathlib import PurePath
from textwrap import dedent
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Mapping, TypeVar, cast
from urllib.parse import quote, urlparse, urlunparse

from litestar import Litestar, MediaType, Request, Response
from litestar.datastructures.cookie import Cookie
from litestar.exceptions import ImproperlyConfiguredException
from litestar.response import Redirect
from litestar.response.base import ASGIResponse
from litestar.serialization import get_serializer
from litestar.status_codes import HTTP_200_OK, HTTP_303_SEE_OTHER, HTTP_307_TEMPORARY_REDIRECT, HTTP_409_CONFLICT
from litestar.utils.deprecation import warn_deprecation
from litestar.utils.empty import value_or_default
from litestar.utils.helpers import get_enum_string_value
from litestar.utils.scope.state import ScopeState
from markupsafe import Markup

from litestar_vite.inertia._utils import get_headers
from litestar_vite.inertia.types import InertiaHeaderType, PageProps
from litestar_vite.plugin import VitePlugin

if TYPE_CHECKING:
    from litestar.app import Litestar
    from litestar.background_tasks import BackgroundTask, BackgroundTasks
    from litestar.connection import ASGIConnection
    from litestar.connection.base import AuthT, StateT, UserT
    from litestar.types import ResponseCookies, ResponseHeaders, TypeEncodersMap

    from litestar_vite.inertia.routes import Routes

    from .plugin import InertiaPlugin

T = TypeVar("T")


def share(
    connection: ASGIConnection[Any, Any, Any, Any],
    key: str,
    value: Any,
) -> None:
    try:
        connection.session.setdefault("_shared", {}).update({key: value})
    except (AttributeError, ImproperlyConfiguredException):
        msg = "Unable to set `share` session state.  A valid session was not found for this request."
        connection.logger.warning(msg)


def error(
    connection: ASGIConnection[Any, Any, Any, Any],
    key: str,
    message: str,
) -> None:
    try:
        connection.session.setdefault("_errors", {}).update({key: message})
    except (AttributeError, ImproperlyConfiguredException):
        msg = "Unable to set `error` session state.  A valid session was not found for this request."
        connection.logger.warning(msg)


def get_shared_props(
    request: ASGIConnection[Any, Any, Any, Any],
    partial_data: set[str] | None = None,
) -> Dict[str, Any]:  # noqa: UP006
    """Return shared session props for a request


    Be sure to call this before `self.create_template_context` if you would like to include the `flash` message details.
    """
    props: dict[str, Any] = {}
    flash: dict[str, list[str]] = defaultdict(list)
    errors: dict[str, Any] = {}
    error_bag = request.headers.get("X-Inertia-Error-Bag", None)
    try:
        errors = request.session.pop("_errors", {})
        props.update(cast("Dict[str,Any]", request.session.pop("_shared", {})))
        for message in cast("List[Dict[str,Any]]", request.session.pop("_messages", [])):
            flash[message["category"]].append(message["message"])

        inertia_plugin = cast("InertiaPlugin", request.app.plugins.get("InertiaPlugin"))
        props.update(inertia_plugin.config.extra_static_page_props)
        for session_prop in inertia_plugin.config.extra_session_page_props:
            if session_prop not in props and session_prop in request.session:
                props[session_prop] = request.session.get(session_prop)

    except (AttributeError, ImproperlyConfiguredException):
        msg = "Unable to generate all shared props.  A valid session was not found for this request."
        request.logger.warning(msg)
    props["flash"] = flash
    props["errors"] = {error_bag: errors} if error_bag is not None else errors
    props["csrf_token"] = value_or_default(ScopeState.from_scope(request.scope).csrf_token, "")
    return props


def js_routes_script(js_routes: Routes) -> Markup:
    @lru_cache
    def _markup_safe_json_dumps(js_routes: str) -> Markup:
        js = js_routes.replace("<", "\\u003c").replace(">", "\\u003e").replace("&", "\\u0026").replace("'", "\\u0027")
        return Markup(js)

    return Markup(
        dedent(f"""
        <script type="module">
        globalThis.routes = JSON.parse('{_markup_safe_json_dumps(js_routes.formatted_routes)}')
        </script>
        """),
    )


class InertiaResponse(Response[T]):
    """Inertia Response"""

    def __init__(
        self,
        content: T,
        *,
        template_name: str | None = None,
        template_str: str | None = None,
        background: BackgroundTask | BackgroundTasks | None = None,
        context: dict[str, Any] | None = None,
        cookies: ResponseCookies | None = None,
        encoding: str = "utf-8",
        headers: ResponseHeaders | None = None,
        media_type: MediaType | str | None = None,
        status_code: int = HTTP_200_OK,
        type_encoders: TypeEncodersMap | None = None,
    ) -> None:
        """Handle the rendering of a given template into a bytes string.

        Args:
            content: A value for the response body that will be rendered into bytes string.
            template_name: Path-like name for the template to be rendered, e.g. ``index.html``.
            template_str: A string representing the template, e.g. ``tmpl = "Hello <strong>World</strong>"``.
            background: A :class:`BackgroundTask <.background_tasks.BackgroundTask>` instance or
                :class:`BackgroundTasks <.background_tasks.BackgroundTasks>` to execute after the response is finished.
                Defaults to ``None``.
            context: A dictionary of key/value pairs to be passed to the temple engine's render method.
            cookies: A list of :class:`Cookie <.datastructures.Cookie>` instances to be set under the response
                ``Set-Cookie`` header.
            encoding: Content encoding
            headers: A string keyed dictionary of response headers. Header keys are insensitive.
            media_type: A string or member of the :class:`MediaType <.enums.MediaType>` enum. If not set, try to infer
                the media type based on the template name. If this fails, fall back to ``text/plain``.
            status_code: A value for the response HTTP status code.
            type_encoders: A mapping of types to callables that transform them into types supported for serialization.
        """
        if template_name and template_str:
            msg = "Either template_name or template_str must be provided, not both."
            raise ValueError(msg)
        self.content = content
        self.background = background
        self.cookies: list[Cookie] = (
            [Cookie(key=key, value=value) for key, value in cookies.items()]
            if isinstance(cookies, Mapping)
            else list(cookies or [])
        )
        self.encoding = encoding
        self.headers: dict[str, Any] = (
            dict(headers) if isinstance(headers, Mapping) else {h.name: h.value for h in headers or {}}
        )
        self.media_type = media_type
        self.status_code = status_code
        self.response_type_encoders = {**(self.type_encoders or {}), **(type_encoders or {})}
        self.context = context or {}
        self.template_name = template_name
        self.template_str = template_str

    def create_template_context(
        self,
        request: Request[UserT, AuthT, StateT],
        page_props: PageProps[T],
        type_encoders: TypeEncodersMap | None = None,
    ) -> dict[str, Any]:
        """Create a context object for the template.

        Args:
            request: A :class:`Request <.connection.Request>` instance.
            page_props: A formatted object to return the inertia configuration.
            type_encoders: A mapping of types to callables that transform them into types supported for serialization.

        Returns:
            A dictionary holding the template context
        """
        csrf_token = value_or_default(ScopeState.from_scope(request.scope).csrf_token, "")
        inertia_props = self.render(page_props, MediaType.JSON, get_serializer(type_encoders)).decode()
        return {
            **self.context,
            "inertia": inertia_props,
            "js_routes": js_routes_script(request.app.state.js_routes),
            "request": request,
            "csrf_input": f'<input type="hidden" name="_csrf_token" value="{csrf_token}" />',
        }

    def to_asgi_response(
        self,
        app: Litestar | None,
        request: Request[UserT, AuthT, StateT],
        *,
        background: BackgroundTask | BackgroundTasks | None = None,
        cookies: Iterable[Cookie] | None = None,
        encoded_headers: Iterable[tuple[bytes, bytes]] | None = None,
        headers: dict[str, str] | None = None,
        is_head_response: bool = False,
        media_type: MediaType | str | None = None,
        status_code: int | None = None,
        type_encoders: TypeEncodersMap | None = None,
    ) -> ASGIResponse:
        if app is not None:
            warn_deprecation(
                version="2.1",
                deprecated_name="app",
                kind="parameter",
                removal_in="3.0.0",
                alternative="request.app",
            )
        inertia_enabled = cast(
            "bool",
            getattr(request, "inertia_enabled", False) or getattr(request, "is_inertia", False),
        )
        is_inertia = cast("bool", getattr(request, "is_inertia", False))
        headers = {**headers, **self.headers} if headers is not None else self.headers
        cookies = self.cookies if cookies is None else itertools.chain(self.cookies, cookies)
        type_encoders = (
            {**type_encoders, **(self.response_type_encoders or {})} if type_encoders else self.response_type_encoders
        )
        if not inertia_enabled:
            media_type = get_enum_string_value(self.media_type or media_type or MediaType.JSON)
            return ASGIResponse(
                background=self.background or background,
                body=self.render(self.content, media_type, get_serializer(type_encoders)),
                cookies=cookies,
                encoded_headers=encoded_headers,
                encoding=self.encoding,
                headers=headers,
                is_head_response=is_head_response,
                media_type=media_type,
                status_code=self.status_code or status_code,
            )
        is_partial_render = cast("bool", getattr(request, "is_partial_render", False))
        partial_keys = cast("set[str]", getattr(request, "partial_keys", {}))
        vite_plugin = request.app.plugins.get(VitePlugin)
        template_engine = vite_plugin.template_config.to_engine()
        headers.update(
            {"Vary": "Accept", **get_headers(InertiaHeaderType(enabled=True))},
        )
        shared_props = get_shared_props(request, partial_data=partial_keys if is_partial_render else None)
        shared_props["content"] = self.content
        page_props = PageProps[T](
            component=request.inertia.route_component,  # type: ignore[attr-defined] # pyright: ignore[reportUnknownArgumentType,reportUnknownMemberType,reportAttributeAccessIssue]
            props=shared_props,  # pyright: ignore[reportArgumentType]
            version=template_engine.asset_loader.version_id,
            url=request.url.path,
        )
        if is_inertia:
            media_type = get_enum_string_value(self.media_type or media_type or MediaType.JSON)
            body = self.render(page_props, media_type, get_serializer(type_encoders))
            return ASGIResponse(
                background=self.background or background,
                body=body,
                cookies=cookies,
                encoded_headers=encoded_headers,
                encoding=self.encoding,
                headers=headers,
                is_head_response=is_head_response,
                media_type=media_type,
                status_code=self.status_code or status_code,
            )

        if not template_engine:
            msg = "Template engine is not configured"
            raise ImproperlyConfiguredException(msg)
        # it should default to HTML at this point unless the user specified something
        media_type = media_type or MediaType.HTML
        if not media_type:
            if self.template_name:
                suffixes = PurePath(self.template_name).suffixes
                for suffix in suffixes:
                    if _type := guess_type(f"name{suffix}")[0]:
                        media_type = _type
                        break
                else:
                    media_type = MediaType.TEXT
            else:
                media_type = MediaType.HTML
        context = self.create_template_context(request, page_props, type_encoders)  # pyright: ignore[reportUnknownMemberType]
        if self.template_str is not None:
            body = template_engine.render_string(self.template_str, context).encode(self.encoding)
        else:
            inertia_plugin = cast("InertiaPlugin", request.app.plugins.get("InertiaPlugin"))
            template_name = self.template_name or inertia_plugin.config.root_template
            # cast to str b/c we know that either template_name cannot be None if template_str is None
            template = template_engine.get_template(template_name)
            body = template.render(**context).encode(self.encoding)

        return ASGIResponse(
            background=self.background or background,
            body=body,
            cookies=cookies,
            encoded_headers=encoded_headers,
            encoding=self.encoding,
            headers=headers,
            is_head_response=is_head_response,
            media_type=media_type,
            status_code=self.status_code or status_code,
        )


class InertiaExternalRedirect(Response[Any]):
    """Client side redirect."""

    def __init__(
        self,
        request: Request[Any, Any, Any],
        redirect_to: str,
        **kwargs: Any,
    ) -> None:
        """Initialize external redirect, Set status code to 409 (required by Inertia),
        and pass redirect url.
        """
        super().__init__(
            content=b"",
            status_code=HTTP_409_CONFLICT,
            headers={"X-Inertia-Location": quote(redirect_to, safe="/#%[]=:;$&()+,!?*@'~")},
            cookies=request.cookies,
            **kwargs,
        )


class InertiaRedirect(Redirect):
    """Client side redirect."""

    def __init__(
        self,
        request: Request[Any, Any, Any],
        redirect_to: str,
        **kwargs: Any,
    ) -> None:
        """Initialize external redirect, Set status code to 409 (required by Inertia),
        and pass redirect url.
        """
        referer = urlparse(request.headers.get("Referer", str(request.base_url)))
        redirect_to = urlunparse(urlparse(redirect_to)._replace(scheme=referer.scheme))
        super().__init__(
            path=redirect_to,
            status_code=HTTP_307_TEMPORARY_REDIRECT if request.method == "GET" else HTTP_303_SEE_OTHER,
            cookies=request.cookies,
            **kwargs,
        )


class InertiaBack(Redirect):
    """Client side redirect."""

    def __init__(
        self,
        request: Request[Any, Any, Any],
        **kwargs: Any,
    ) -> None:
        """Initialize external redirect, Set status code to 409 (required by Inertia),
        and pass redirect url.
        """
        super().__init__(
            path=request.headers.get("Referer", str(request.base_url)),
            status_code=HTTP_307_TEMPORARY_REDIRECT if request.method == "GET" else HTTP_303_SEE_OTHER,
            cookies=request.cookies,
            **kwargs,
        )



"""Metadata for the Project."""

from __future__ import annotations

from importlib.metadata import PackageNotFoundError, metadata, version

__all__ = ("__project__", "__version__")

try:
    __version__ = version("litestar_vite")
    """Version of the project."""
    __project__ = metadata("litestar_vite")["Name"]
    """Name of the project."""
except PackageNotFoundError:  # pragma: no cover
    __version__ = "0.0.0"
    __project__ = "Litestar Vite"
finally:
    del version, PackageNotFoundError, metadata



from __future__ import annotations

from litestar_vite import inertia
from litestar_vite.config import ViteConfig, ViteTemplateConfig
from litestar_vite.loader import ViteAssetLoader
from litestar_vite.plugin import VitePlugin
from litestar_vite.template_engine import ViteTemplateEngine

__all__ = ("ViteAssetLoader", "ViteConfig", "VitePlugin", "ViteTemplateConfig", "ViteTemplateEngine", "inertia")



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING, Any, Callable

if TYPE_CHECKING:
    from litestar_vite.inertia.types import InertiaHeaderType


class InertiaHeaders(str, Enum):
    """Enum for Inertia Headers"""

    ENABLED = "X-Inertia"
    VERSION = "X-Inertia-Version"
    PARTIAL_DATA = "X-Inertia-Partial-Data"
    PARTIAL_COMPONENT = "X-Inertia-Partial-Component"
    LOCATION = "X-Inertia-Location"
    REFERER = "Referer"


def get_enabled_header(enabled: bool = True) -> dict[str, Any]:
    """True if inertia is enabled."""

    return {InertiaHeaders.ENABLED.value: "true" if enabled else "false"}


def get_version_header(version: str) -> dict[str, Any]:
    """Return headers for change swap method response."""
    return {InertiaHeaders.VERSION.value: version}


def get_partial_data_header(partial: str) -> dict[str, Any]:
    """Return headers for a partial data response."""
    return {InertiaHeaders.PARTIAL_DATA.value: partial}


def get_partial_component_header(partial: str) -> dict[str, Any]:
    """Return headers for a partial data response."""
    return {InertiaHeaders.PARTIAL_COMPONENT.value: partial}


def get_headers(inertia_headers: InertiaHeaderType) -> dict[str, Any]:
    """Return headers for Inertia responses."""
    if not inertia_headers:
        msg = "Value for inertia_headers cannot be None."
        raise ValueError(msg)
    inertia_headers_dict: dict[str, Callable[..., dict[str, Any]]] = {
        "enabled": get_enabled_header,
        "partial_data": get_partial_data_header,
        "partial_component": get_partial_component_header,
        "version": get_version_header,
    }

    header: dict[str, Any] = {}
    response: dict[str, Any]
    key: str
    value: Any

    for key, value in inertia_headers.items():
        if value is not None:
            response = inertia_headers_dict[key](value)
            header.update(response)
    return header



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

from click import Context, group, option
from click import Path as ClickPath
from litestar.cli._utils import (
    LitestarEnv,
    LitestarGroup,
)

if TYPE_CHECKING:
    from litestar import Litestar


@group(cls=LitestarGroup, name="assets")
def vite_group() -> None:
    """Manage Vite Tasks."""


@vite_group.command(
    name="init",
    help="Initialize vite for your project.",
)
@option(
    "--root-path",
    type=ClickPath(dir_okay=True, file_okay=False, path_type=Path),
    help="The path for the initial the Vite application.  This is the equivalent to the top-level folder in a normal Vue or React application..",
    default=None,
    required=False,
)
@option(
    "--bundle-path",
    type=ClickPath(dir_okay=True, file_okay=False, path_type=Path),
    help="The path for the built Vite assets.  This is the where the output of `npm run build` will write files.",
    default=None,
    required=False,
)
@option(
    "--resource-path",
    type=ClickPath(dir_okay=True, file_okay=False, path_type=Path),
    help="The path to your Javascript/Typescript source and associated assets.  If this were a standalone Vue or React app, this would point to your `src/` folder.",
    default=None,
    required=False,
)
@option(
    "--public-path",
    type=ClickPath(dir_okay=True, file_okay=False, path_type=Path),
    help="The optional path to your public/static JS assets.  If this were a standalone Vue or React app, this would point to your `public/` folder.",
    default=None,
    required=False,
)
@option("--asset-url", type=str, help="Base url to serve assets from.", default=None, required=False)
@option(
    "--vite-port",
    type=int,
    help="The port to run the vite server against.",
    default=None,
    is_flag=True,
    required=False,
)
@option(
    "--enable-ssr",
    type=bool,
    help="Enable SSR Support.",
    required=False,
    show_default=False,
    is_flag=True,
)
@option("--overwrite", type=bool, help="Overwrite any files in place.", default=False, is_flag=True)
@option("--verbose", type=bool, help="Enable verbose output.", default=False, is_flag=True)
@option(
    "--no-prompt",
    help="Do not prompt for confirmation and use all defaults for initializing the project.",
    type=bool,
    default=False,
    required=False,
    show_default=True,
    is_flag=True,
)
@option(
    "--no-install",
    help="Do not execute the installation commands after generating templates.",
    type=bool,
    default=False,
    required=False,
    show_default=True,
    is_flag=True,
)
def vite_init(
    ctx: Context,
    vite_port: int | None,
    enable_ssr: bool | None,
    asset_url: str | None,
    root_path: Path | None,
    bundle_path: Path | None,
    resource_path: Path | None,
    public_path: Path | None,
    overwrite: bool,
    verbose: bool,
    no_prompt: bool,
    no_install: bool,
) -> None:  # sourcery skip: low-code-quality
    """Run vite build."""
    import os
    import sys
    from importlib.util import find_spec
    from pathlib import Path

    from litestar.cli._utils import (
        console,
    )
    from rich.prompt import Confirm

    from litestar_vite.commands import execute_command, init_vite
    from litestar_vite.plugin import VitePlugin

    if callable(ctx.obj):
        ctx.obj = ctx.obj()
    elif verbose:
        ctx.obj.app.debug = True
    env: LitestarEnv = ctx.obj
    plugin = env.app.plugins.get(VitePlugin)
    config = plugin._config  # pyright: ignore[reportPrivateUsage]

    console.rule("[yellow]Initializing Vite[/]", align="left")
    resource_path = Path(resource_path or config.resource_dir)
    public_path = Path(public_path or config.public_dir)
    bundle_path = Path(bundle_path or config.bundle_dir)
    enable_ssr = enable_ssr or config.ssr_enabled
    asset_url = asset_url or config.asset_url
    vite_port = vite_port or config.port
    hot_file = Path(bundle_path / config.hot_file)
    root_path = Path(root_path or config.root_dir or resource_path.parent)

    if any(output_path.exists() for output_path in (bundle_path, resource_path)) and not any(
        [overwrite, no_prompt],
    ):
        confirm_overwrite = Confirm.ask(
            "Files were found in the paths specified.  Are you sure you wish to overwrite the contents?",
        )
        if not confirm_overwrite:
            console.print("Skipping Vite initialization")
            sys.exit(2)
    for output_path in (bundle_path, resource_path, root_path):
        if output_path.exists():
            console.print(f" * Creating {output_path!s}")
        output_path.mkdir(parents=True, exist_ok=True)

    enable_ssr = (
        True
        if enable_ssr
        else False
        if no_prompt
        else Confirm.ask(
            "Do you intend to use Litestar with any SSR framework?",
        )
    )
    init_vite(
        app=env.app,
        root_path=root_path,
        enable_ssr=enable_ssr,
        vite_port=vite_port,
        asset_url=asset_url,
        resource_path=resource_path,
        public_path=public_path,
        bundle_path=bundle_path,
        hot_file=hot_file,
        litestar_port=env.port or 8000,
    )
    if not no_install:
        if find_spec("nodeenv") is not None and plugin.config.detect_nodeenv:
            """Detect nodeenv installed in the current python env before using a global version"""
            nodeenv_command = (
                str(Path(Path(sys.executable) / "nodeenv"))
                if Path(Path(sys.executable) / "nodeenv").exists()
                else "nodeenv"
            )
            install_dir = os.environ.get("VIRTUAL_ENV", sys.prefix)
            console.rule("[yellow]Starting Nodeenv installation process[/]", align="left")
            console.print(f"Installing Node environment into {install_dir}")
            execute_command(command_to_run=[nodeenv_command, install_dir, "--force", "--quiet"], cwd=root_path)

        console.rule("[yellow]Starting package installation process[/]", align="left")

        execute_command(command_to_run=plugin.config.install_command, cwd=root_path)


@vite_group.command(
    name="install",
    help="Install frontend packages.",
)
@option("--verbose", type=bool, help="Enable verbose output.", default=False, is_flag=True)
def vite_install(app: Litestar, verbose: bool) -> None:
    """Run vite build."""
    import os
    import sys
    from importlib.util import find_spec
    from pathlib import Path

    from litestar.cli._utils import console

    from litestar_vite.commands import execute_command
    from litestar_vite.plugin import VitePlugin

    if verbose:
        app.debug = True
    plugin = app.plugins.get(VitePlugin)

    if find_spec("nodeenv") is not None and plugin.config.detect_nodeenv:
        """Detect nodeenv installed in the current python env before using a global version"""
        nodeenv_command = (
            str(Path(Path(sys.executable) / "nodeenv"))
            if Path(Path(sys.executable) / "nodeenv").exists()
            else "nodeenv"
        )
        install_dir = os.environ.get("VIRTUAL_ENV", sys.prefix)
        console.rule("[yellow]Starting Nodeenv installation process[/]", align="left")
        console.print("Installing Node environment to %s:", install_dir)
        execute_command(command_to_run=[nodeenv_command, install_dir, "--force", "--quiet"], cwd=plugin.config.root_dir)

    console.rule("[yellow]Starting package installation process[/]", align="left")

    execute_command(command_to_run=plugin.config.install_command, cwd=plugin.config.root_dir)


@vite_group.command(
    name="build",
    help="Building frontend assets with Vite.",
)
@option("--verbose", type=bool, help="Enable verbose output.", default=False, is_flag=True)
def vite_build(app: Litestar, verbose: bool) -> None:
    """Run vite build."""
    from litestar.cli._utils import (
        console,
    )

    from litestar_vite.commands import execute_command
    from litestar_vite.plugin import VitePlugin, set_environment

    if verbose:
        app.debug = True
    console.rule("[yellow]Starting Vite build process[/]", align="left")
    plugin = app.plugins.get(VitePlugin)
    if plugin.config.set_environment:
        set_environment(config=plugin.config)
    p = execute_command(command_to_run=plugin.config.build_command, cwd=plugin.config.root_dir)
    if p.returncode == 0:
        console.print("[bold green] Assets built.[/]")
    else:
        console.print("[bold red] There was an error building the assets.[/]")


@vite_group.command(
    name="serve",
    help="Serving frontend assets with Vite.",
)
@option("--verbose", type=bool, help="Enable verbose output.", default=False, is_flag=True)
def vite_serve(app: Litestar, verbose: bool) -> None:
    """Run vite serve."""
    from litestar.cli._utils import (
        console,
    )

    from litestar_vite.commands import execute_command
    from litestar_vite.plugin import VitePlugin, set_environment

    if verbose:
        app.debug = True

    plugin = app.plugins.get(VitePlugin)
    if plugin.config.set_environment:
        set_environment(config=plugin.config)
    if plugin.config.hot_reload:
        console.rule("[yellow]Starting Vite process with HMR Enabled[/]", align="left")
    else:
        console.rule("[yellow]Starting Vite watch and build process[/]", align="left")
    command_to_run = plugin.config.run_command if plugin.config.hot_reload else plugin.config.build_watch_command
    execute_command(command_to_run=command_to_run, cwd=plugin.config.root_dir)
    console.print("[yellow]Vite process stopped.[/]")


@vite_group.command(
    name="generate-routes",
    help="Generate a JSON file with the route configuration",
)
@option(
    "--output",
    help="output file path",
    type=ClickPath(dir_okay=False, path_type=Path),
    default=Path("routes.json"),
    show_default=True,
)
@option("--verbose", type=bool, help="Enable verbose output.", default=False, is_flag=True)
def generate_js_routes(app: Litestar, output: Path, verbose: bool) -> None:
    """Run vite serve."""
    import msgspec
    from litestar.cli._utils import (
        LitestarCLIException,
        console,
    )
    from litestar.serialization import encode_json, get_serializer

    from litestar_vite.plugin import VitePlugin, set_environment

    if verbose:
        app.debug = True
    serializer = get_serializer(app.type_encoders)
    plugin = app.plugins.get(VitePlugin)
    if plugin.config.set_environment:
        set_environment(config=plugin.config)
    content = msgspec.json.format(
        encode_json(app.openapi_schema.to_schema(), serializer=serializer),
        indent=4,
    )

    try:
        output.write_bytes(content)
    except OSError as e:  # pragma: no cover
        msg = f"failed to write schema to path {output}"
        raise LitestarCLIException(msg) from e

    console.print("[yellow]Vite process stopped.[/]")



from __future__ import annotations

import json
from functools import cached_property
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Any, ClassVar
from urllib.parse import urljoin

if TYPE_CHECKING:
    from litestar_vite.config import ViteConfig


class ViteAssetLoader:
    """Vite  manifest loader.

    Please see: https://vitejs.dev/guide/backend-integration.html
    """

    _instance: ClassVar[ViteAssetLoader | None] = None

    def __init__(self, config: ViteConfig) -> None:
        self._config = config
        self._manifest: dict[str, Any] = {}
        self._manifest_content: str = ""
        self._vite_base_path: str | None = None

    @classmethod
    def initialize_loader(cls, config: ViteConfig) -> ViteAssetLoader:
        """Singleton manifest loader."""
        if cls._instance is None:
            cls._instance = cls(config=config)
            cls._instance.parse_manifest()
        return cls._instance

    @cached_property
    def version_id(self) -> str:
        if self._manifest_content != "":
            return str(hash(self.manifest_content))
        return "1.0"

    def parse_manifest(self) -> None:
        """Parse the Vite manifest file.

        The manifest file is a JSON file that maps source files to their corresponding output files.
        Example manifest file structure:

        .. code-block:: json

            {
                "main.js": {
                    "file": "assets/main.4889e940.js",
                    "src": "main.js",
                    "isEntry": true,
                    "dynamicImports": ["views/foo.js"],
                    "css": ["assets/main.b82dbe22.css"],
                    "assets": ["assets/asset.0ab0f9cd.png"]
                },
                "views/foo.js": {
                    "file": "assets/foo.869aea0d.js",
                    "src": "views/foo.js",
                    "isDynamicEntry": true,
                    "imports": ["_shared.83069a53.js"]
                },
                "_shared.83069a53.js": {
                    "file": "assets/shared.83069a53.js"
                }
            }

        The manifest is parsed and stored in memory for asset resolution during template rendering.
        """
        if self._config.hot_reload and self._config.dev_mode:
            hot_file_path = Path(
                f"{self._config.bundle_dir}/{self._config.hot_file}",
            )
            if hot_file_path.exists():
                with hot_file_path.open() as hot_file:
                    self._vite_base_path = hot_file.read()

        else:
            manifest_path = Path(f"{self._config.bundle_dir}/{self._config.manifest_name}")
            try:
                if manifest_path.exists():
                    with manifest_path.open() as manifest_file:
                        self.manifest_content = manifest_file.read()
                        self._manifest = json.loads(self.manifest_content)
                else:
                    self._manifest = {}
            except Exception as exc:
                msg = "There was an issue reading the Vite manifest file at  %s. Did you forget to build your assets?"
                raise RuntimeError(
                    msg,
                    manifest_path,
                ) from exc

    def generate_ws_client_tags(self) -> str:
        """Generate the script tag for the Vite WS client for HMR.

        Only used when hot module reloading is enabled, in production this method returns an empty string.

        Returns:
            str: The script tag or an empty string.
        """
        if self._config.hot_reload and self._config.dev_mode:
            return self._script_tag(
                self._vite_server_url("@vite/client"),
                {"type": "module"},
            )
        return ""

    def generate_react_hmr_tags(self) -> str:
        """Generate the script tag for the Vite WS client for HMR.

        Only used when hot module reloading is enabled, in production this method returns an empty string.

        Returns:
            str: The script tag or an empty string.
        """
        if self._config.is_react and self._config.hot_reload and self._config.dev_mode:
            return dedent(f"""
                <script type="module">
                import RefreshRuntime from '{self._vite_server_url()}@react-refresh'
                RefreshRuntime.injectIntoGlobalHook(window)
                window.$RefreshReg$ = () => {{}}
                window.$RefreshSig$ = () => (type) => type
                window.__vite_plugin_react_preamble_installed__=true
                </script>
                """)
        return ""

    def generate_asset_tags(self, path: str | list[str], scripts_attrs: dict[str, str] | None = None) -> str:
        """Generate all assets include tags for the file in argument.

        Returns:
            str: All tags to import this asset in your HTML page.
        """
        if isinstance(path, str):
            path = [path]
        if self._config.hot_reload and self._config.dev_mode:
            return "".join(
                [
                    self._style_tag(self._vite_server_url(p))
                    if p.endswith(".css")
                    else self._script_tag(
                        self._vite_server_url(p),
                        {"type": "module", "async": "", "defer": ""},
                    )
                    for p in path
                ],
            )

        if any(p for p in path if p not in self._manifest):
            msg = "Cannot find %s in Vite manifest at %s.  Did you forget to build your assets after an update?"
            raise RuntimeError(
                msg,
                path,
                Path(f"{self._config.bundle_dir}/{self._config.manifest_name}"),
            )

        tags: list[str] = []
        manifest_entry: dict[str, Any] = {}
        manifest_entry.update({p: self._manifest[p] for p in path})
        if not scripts_attrs:
            scripts_attrs = {"type": "module", "async": "", "defer": ""}
        for manifest in manifest_entry.values():
            if "css" in manifest:
                tags.extend(
                    self._style_tag(urljoin(self._config.asset_url, css_path)) for css_path in manifest.get("css", {})
                )
            # Add dependent "vendor"
            if "imports" in manifest:
                tags.extend(
                    self.generate_asset_tags(vendor_path, scripts_attrs=scripts_attrs)
                    for vendor_path in manifest.get("imports", {})
                )
            # Add the script by itself
            if manifest.get("file").endswith(".css"):
                tags.append(
                    self._style_tag(urljoin(self._config.asset_url, manifest["file"])),
                )
            else:
                tags.append(
                    self._script_tag(
                        urljoin(self._config.asset_url, manifest["file"]),
                        attrs=scripts_attrs,
                    ),
                )
        return "".join(tags)

    def _vite_server_url(self, path: str | None = None) -> str:
        """Generate an URL to and asset served by the Vite development server.

        Keyword Arguments:
            path: Path to the asset. (default: {None})

        Returns:
            str: Full URL to the asset.
        """
        base_path = self._vite_base_path or f"{self._config.protocol}://{self._config.host}:{self._config.port}"
        return urljoin(
            base_path,
            urljoin(self._config.asset_url, path if path is not None else ""),
        )

    def _script_tag(self, src: str, attrs: dict[str, str] | None = None) -> str:
        """Generate an HTML script tag."""
        if attrs is None:
            attrs = {}
        attrs_str = " ".join([f'{key}="{value}"' for key, value in attrs.items()])
        return f'<script {attrs_str} src="{src}"></script>'

    def _style_tag(self, href: str) -> str:
        """Generate and HTML <link> stylesheet tag for CSS.

        Args:
            href: CSS file URL.

        Returns:
            str: CSS link tag.
        """
        return f'<link rel="stylesheet" href="{href}" />'



from typing import Any

import pytest
from litestar import MediaType, Request, get
from litestar.middleware.session.server_side import ServerSideSessionConfig
from litestar.status_codes import HTTP_200_OK
from litestar.stores.memory import MemoryStore
from litestar.testing import create_test_client  # pyright: ignore[reportUnknownVariableType]

from litestar_vite.inertia import InertiaHeaders, InertiaPlugin, InertiaRequest
from litestar_vite.plugin import VitePlugin

pytestmark = pytest.mark.anyio


def test_health_check(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/health-check", media_type=MediaType.TEXT)
    async def health_check() -> str:
        return "healthy"

    with create_test_client(
        route_handlers=health_check,
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/health-check")
        assert response.status_code == HTTP_200_OK
        assert response.text == "healthy"


async def test_is_inertia_default(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/")
    async def handler(request: InertiaRequest[Any, Any, Any]) -> bool:
        return bool(request.is_inertia)

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/")
        assert response.text == "false"


async def test_is_inertia_false(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/")
    async def handler(request: InertiaRequest[Any, Any, Any]) -> bool:
        return bool(request.is_inertia)

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/", headers={InertiaHeaders.ENABLED.value: "false"})
        assert response.text == "false"


async def test_is_inertia_true(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/")
    async def handler(request: InertiaRequest[Any, Any, Any]) -> bool:
        return bool(request.is_inertia)

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/", headers={InertiaHeaders.ENABLED.value: "true"})
        assert (
            response.text
            == '{"component":null,"url":"/","version":"1.0","props":{"flash":{},"errors":{},"csrf_token":"","content":true}}'
        )


async def test_component_prop_default(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/")
    async def handler(request: InertiaRequest[Any, Any, Any]) -> bool:
        return request.inertia_enabled

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/")
        assert response.text == "false"


async def test_component_enabled(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/", component="Home")
    async def handler(request: InertiaRequest[Any, Any, Any]) -> bool:
        return request.inertia_enabled

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/")
        assert response.text.startswith("<!DOCTYPE html>")


async def test_default_route_no_component(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/")
    async def handler(request: Request[Any, Any, Any]) -> str:
        return request.inertia.route_component or ""  # pyright: ignore

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/")
        assert response.text == ""



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from litestar import Request
from litestar.middleware import AbstractMiddleware
from litestar.types import Receive, Scope, Send

from litestar_vite.inertia.response import InertiaRedirect
from litestar_vite.plugin import VitePlugin

if TYPE_CHECKING:
    from litestar.connection.base import (
        AuthT,
        StateT,
        UserT,
    )
    from litestar.types import ASGIApp, Receive, Scope, Send


async def redirect_on_asset_version_mismatch(request: Request[UserT, AuthT, StateT]) -> InertiaRedirect | None:
    if getattr(request, "is_inertia", None) is None:
        return None
    inertia_version = request.headers.get("X-Inertia-Version")
    if inertia_version is None:
        return None

    vite_plugin = request.app.plugins.get(VitePlugin)
    template_engine = vite_plugin.template_config.to_engine()
    if inertia_version == template_engine.asset_loader.version_id:
        return None
    return InertiaRedirect(request, redirect_to=str(request.url))


class InertiaMiddleware(AbstractMiddleware):
    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)
        self.app = app

    async def __call__(
        self,
        scope: "Scope",
        receive: "Receive",
        send: "Send",
    ) -> None:
        request = Request[Any, Any, Any](scope=scope)
        redirect = await redirect_on_asset_version_mismatch(request)
        if redirect is not None:
            response = redirect.to_asgi_response(app=None, request=request)  # pyright: ignore[reportUnknownMemberType]
            await response(scope, receive, send)
        else:
            await self.app(scope, receive, send)



from __future__ import annotations

from pathlib import Path

import pytest

from litestar_vite.config import ViteConfig

pytestmark = pytest.mark.anyio
here = Path(__file__).parent


@pytest.fixture
def anyio_backend() -> str:
    return "asyncio"


# Define a fixture for ViteConfig
@pytest.fixture
def vite_config() -> ViteConfig:
    # Mock the ViteConfig with necessary attributes for testing
    return ViteConfig(
        bundle_dir=Path(here / "test_app" / "web" / "public"),
        resource_dir=Path(here / "test_app" / "web" / "resources"),
        template_dir=Path(here / "templates"),
        hot_reload=True,
    )



from __future__ import annotations

import os
from dataclasses import dataclass, field
from functools import cached_property
from inspect import isclass
from pathlib import Path
from typing import TYPE_CHECKING, cast

from litestar.exceptions import ImproperlyConfiguredException
from litestar.template import TemplateConfig
from litestar.template.config import EngineType

if TYPE_CHECKING:
    from collections.abc import Callable

    from litestar.types import PathType

__all__ = ("ViteConfig", "ViteTemplateConfig")
TRUE_VALUES = {"True", "true", "1", "yes", "Y", "T"}


@dataclass
class ViteConfig:
    """Configuration for ViteJS support.

    To enable Vite integration, pass an instance of this class to the
    :class:`Litestar <litestar.app.Litestar>` constructor using the
    'plugins' key.
    """

    bundle_dir: Path | str = field(default="public")
    """Location of the compiled assets from  Vite.

    The manifest file will also be found here.
    """
    resource_dir: Path | str = field(default="resources")
    """The directory where all typescript/javascript source are written.

    In a standalone Vue or React application, this would be equivalent to the ``./src`` directory.
    """
    template_dir: Path | str | None = field(default="templates")
    """Location of the Jinja2 template file."""
    public_dir: Path | str = field(default="public")
    """The optional public directory Vite serves assets from.

    In a standalone Vue or React application, this would be equivalent to the ``./public`` directory.
    """
    manifest_name: str = "manifest.json"
    """Name of the manifest file."""
    hot_file: str = "hot"
    """Name of the hot file.

    This file contains a single line containing the host, protocol, and port the Vite server is running.
    """
    hot_reload: bool = field(
        default_factory=lambda: os.getenv("VITE_HOT_RELOAD", "True") in TRUE_VALUES,
    )
    """Enable HMR for Vite development server."""
    ssr_enabled: bool = False
    """Enable SSR."""
    ssr_output_dir: Path | str | None = None
    """SSR Output path"""
    root_dir: Path | str | None = None
    """The is the base path to your application.

   In a standalone Vue or React application, this would be equivalent to the top-level project folder containing the ``./src`` directory.

    """
    is_react: bool = False
    """Enable React components."""
    asset_url: str = field(default_factory=lambda: os.getenv("ASSET_URL", "/static/"))
    """Base URL to generate for static asset references.

    This URL will be prepended to anything generated from Vite.
    """
    host: str = field(default_factory=lambda: os.getenv("VITE_HOST", "localhost"))
    """Default host to use for Vite server."""
    protocol: str = "http"
    """Protocol to use for communication"""
    port: int = field(default_factory=lambda: int(os.getenv("VITE_PORT", "5173")))
    """Default port to use for Vite server."""
    run_command: list[str] = field(default_factory=lambda: ["npm", "run", "dev"])
    """Default command to use for running Vite."""
    build_watch_command: list[str] = field(default_factory=lambda: ["npm", "run", "watch"])
    """Default command to use for dev building with Vite."""
    build_command: list[str] = field(default_factory=lambda: ["npm", "run", "build"])
    """Default command to use for building with Vite."""
    install_command: list[str] = field(default_factory=lambda: ["npm", "install"])
    """Default command to use for installing Vite."""
    use_server_lifespan: bool = field(
        default_factory=lambda: os.getenv("VITE_USE_SERVER_LIFESPAN", "False") in TRUE_VALUES,
    )
    """Utilize the server lifespan hook to run Vite."""
    dev_mode: bool = field(
        default_factory=lambda: os.getenv("VITE_DEV_MODE", "False") in TRUE_VALUES,
    )
    """When True, Vite will run with HMR or watch build"""
    detect_nodeenv: bool = True
    """When True, The initializer will install and configure nodeenv if present"""
    set_environment: bool = True
    """When True, configuration in this class will be set into environment variables.

    This can be useful to ensure Vite always uses the configuration supplied to the plugin
    """
    set_static_folders: bool = True
    """When True, Litestar will automatically serve assets at the `ASSET_URL` path.
    """

    def __post_init__(self) -> None:
        """Ensure that directory is set if engine is a class."""
        if self.root_dir is not None and isinstance(self.root_dir, str):
            self.root_dir = Path(self.root_dir)
        if self.template_dir is not None and isinstance(self.template_dir, str):
            self.template_dir = Path(self.template_dir)
        if self.public_dir and isinstance(self.public_dir, str):
            self.public_dir = Path(self.public_dir)
        if isinstance(self.resource_dir, str):
            self.resource_dir = Path(self.resource_dir)
        if isinstance(self.bundle_dir, str):
            self.bundle_dir = Path(self.bundle_dir)
        if isinstance(self.ssr_output_dir, str):
            self.ssr_output_dir = Path(self.ssr_output_dir)


@dataclass
class ViteTemplateConfig(TemplateConfig[EngineType]):
    """Configuration for Templating.

    To enable templating, pass an instance of this class to the
    :class:`Litestar <litestar.app.Litestar>` constructor using the
    'template_config' key.
    """

    config: ViteConfig = field(default_factory=lambda: ViteConfig())
    """A a config for the vite engine`."""
    engine: type[EngineType] | EngineType | None = field(default=None)
    """A template engine adhering to the :class:`TemplateEngineProtocol <litestar.template.TemplateEngineProtocol>`."""
    directory: PathType | list[PathType] | None = field(default=None)
    """A directory or list of directories from which to serve templates."""
    engine_callback: Callable[[EngineType], None] | None = field(default=None)
    """A callback function that allows modifying the instantiated templating
    protocol."""

    instance: EngineType | None = field(default=None)
    """An instance of the templating protocol."""

    def __post_init__(self) -> None:
        """Ensure that directory is set if engine is a class."""
        if isclass(self.engine) and not self.directory:  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
            msg = "directory is a required kwarg when passing a template engine class"
            raise ImproperlyConfiguredException(msg)
        """Ensure that directory is not set if instance is."""
        if self.instance is not None and self.directory is not None:  # pyright: ignore[reportUnknownMemberType]
            msg = "directory cannot be set if instance is"
            raise ImproperlyConfiguredException(msg)

    def to_engine(self) -> EngineType:
        """Instantiate the template engine."""
        template_engine = cast(
            "EngineType",
            self.engine(directory=self.directory, config=self.config, engine_instance=None)  # pyright: ignore[reportUnknownMemberType,reportCallIssue]
            if isclass(self.engine)  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
            else self.engine,  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
        )
        if callable(self.engine_callback):
            self.engine_callback(template_engine)  # pyright: ignore[reportUnknownMemberType,reportUnknownArgumentType]
        return template_engine

    @cached_property
    def engine_instance(self) -> EngineType:
        """Return the template engine instance."""
        return self.to_engine() if self.instance is None else self.instance



from __future__ import annotations

from pathlib import Path

from litestar import Litestar

from litestar_vite import ViteConfig, VitePlugin

here = Path(__file__).parent

vite = VitePlugin(
    config=ViteConfig(
        bundle_dir=Path(here / "web" / "public"),
        resource_dir=Path(here / "web" / "resources"),
        template_dir=Path(here / "web" / "templates"),
        hot_reload=True,
    ),
)
app = Litestar(plugins=[vite])



import textwrap
from pathlib import Path

from click.testing import CliRunner
from litestar.cli._utils import LitestarGroup

from tests.test_cli.conftest import CreateAppFileFixture


def test_basic_command(runner: CliRunner, create_app_file: CreateAppFileFixture, root_command: LitestarGroup) -> None:
    template_dir = Path(Path(__file__).parent.parent / "templates")
    app_file_content = textwrap.dedent(
        f"""
from __future__ import annotations

from pathlib import Path

from litestar import Controller, Litestar, get
from litestar.response import Template

from litestar_vite import ViteConfig, VitePlugin


class WebController(Controller):
    include_in_schema = False

    @get("/")
    async def index(self) -> Template:
        return Template(template_name="index.html.j2")

vite = VitePlugin(config=ViteConfig(template_dir='{template_dir!s}'))

app = Litestar(plugins=[vite])
    """,
    )
    app_file = create_app_file("command_test_app.py", content=app_file_content)
    result = runner.invoke(root_command, ["--app", f"{app_file.stem}:app", "assets"])

    assert not result.exception
    assert "Manage Vite Tasks." in result.output


def test_init_command_no_prompt(
    runner: CliRunner,
    create_app_file: CreateAppFileFixture,
    root_command: LitestarGroup,
    tmp_project_dir: Path,
) -> None:
    template_dir = Path(Path(__file__).parent.parent / "templates")
    app_file_content = textwrap.dedent(
        f"""
from __future__ import annotations

from pathlib import Path

from litestar import Controller, Litestar, get
from litestar.response import Template

from litestar_vite import ViteConfig, VitePlugin


class WebController(Controller):
    include_in_schema = False

    @get("/")
    async def index(self) -> Template:
        return Template(template_name="index.html.j2")

vite = VitePlugin(config=ViteConfig(template_dir='{template_dir!s}'))

app = Litestar(plugins=[vite])
    """,
    )
    app_file = create_app_file("command_test_app.py", content=app_file_content)
    result = runner.invoke(
        root_command,
        ["--app-dir", f"{app_file.parent!s}", "--app", f"{app_file.stem}:app", "assets", "init", "--no-prompt"],
    )

    assert "Initializing Vite" in result.output
    assert Path(Path(tmp_project_dir) / "vite.config.ts").exists()
    assert Path(Path(tmp_project_dir) / "package.json").exists()
    assert Path(Path(tmp_project_dir) / "tsconfig.json").exists()
    assert Path(Path(tmp_project_dir) / "resources" / "main.ts").exists()
    assert Path(Path(tmp_project_dir) / "resources" / "styles.css").exists()


def test_init_install_build(
    runner: CliRunner,
    create_app_file: CreateAppFileFixture,
    root_command: LitestarGroup,
    tmp_project_dir: Path,
) -> None:
    app_file_content = textwrap.dedent(
        """
from __future__ import annotations

from litestar import Litestar

from litestar_vite import VitePlugin

vite = VitePlugin()

app = Litestar(plugins=[vite])
    """,
    )
    app_file = create_app_file("command_test_app.py", content=app_file_content)
    _ = runner.invoke(root_command, ["--app", f"{app_file.stem}:app", "assets", "init", "--no-prompt"])
    _ = runner.invoke(root_command, ["--app", f"{app_file.stem}:app", "assets", "install"])
    _ = runner.invoke(root_command, ["--app", f"{app_file.stem}:app", "assets", "build"])
    assert Path(app_file.parent / "vite.config.ts").exists()
    assert Path(app_file.parent / "package.json").exists()
    assert Path(app_file.parent / "tsconfig.json").exists()
    assert Path(app_file.parent / "resources" / "main.ts").exists()
    assert Path(app_file.parent / "resources" / "styles.css").exists()



from __future__ import annotations

from pathlib import Path

import pytest
from jinja2 import Environment, FileSystemLoader, select_autoescape

from litestar_vite.commands import VITE_INIT_TEMPLATES, get_template

pytestmark = pytest.mark.anyio
here = Path(__file__).parent


@pytest.fixture
def vite_template_env() -> Environment:
    return Environment(
        loader=FileSystemLoader([Path(here.parent / "litestar_vite" / "templates")]),
        autoescape=select_autoescape(),
    )


def test_get_template(vite_template_env: Environment) -> None:
    init_templates = {
        template_name: get_template(environment=vite_template_env, name=template_name)
        for template_name in VITE_INIT_TEMPLATES
    }
    assert len(init_templates.keys()) == len(VITE_INIT_TEMPLATES)



from __future__ import annotations

from typing import Any

import pytest
from jinja2 import Environment, FileSystemLoader

from litestar_vite.config import ViteConfig
from litestar_vite.template_engine import ViteTemplateEngine


# Define a fixture for Jinja Environment
@pytest.fixture
def jinja_env(vite_config: ViteConfig) -> Environment:
    # Mock the Jinja Environment with necessary attributes for testing
    if vite_config.template_dir is None:
        pytest.skip(reason="Template directory is required for testing Jinja.  Skipping.")
    loader = FileSystemLoader(searchpath=vite_config.template_dir)
    return Environment(loader=loader, autoescape=True)


# Define a fixture for ViteTemplateEngine
@pytest.fixture
def vite_template_engine(vite_config: ViteConfig, jinja_env: Environment) -> ViteTemplateEngine:
    # Instantiate the ViteTemplateEngine with mock objects
    return ViteTemplateEngine.from_environment(config=vite_config, jinja_environment=jinja_env)


# Happy path tests for get_hmr_client
@pytest.mark.parametrize(
    "expected_output, test_id",
    [
        # Test with a context that should return a non-empty script tag
        # ("<script>...</script>", "dev_mode"),
        # Test with a context that should return an empty string
        ("", "run_mode"),
    ],
)
def test_get_hmr_client(
    vite_template_engine: ViteTemplateEngine,
    expected_output: str,
    test_id: str,
) -> None:
    # Act
    result = vite_template_engine.get_hmr_client()

    # Assert
    assert str(result) == expected_output, f"Failed test ID: {test_id}"


# Happy path tests for get_asset_tag
@pytest.mark.parametrize(
    "path, scripts_attrs, expected_output, test_id",
    [
        # Test with realistic values for a single asset path
        (
            "resources/main.ts",
            None,
            '<script type="module" async="" defer="" src="/static/assets/main-l0sNRNKZ.js"></script>',
            "asset_single",
        ),
        # Test with additional script attributes
        (
            "resources/styles.css",
            {"async": "true"},
            '<script async="true" src="/static/assets/styles-l0sNRNKZ.js"></script>',
            "asset_attrs",
        ),
        # Test with realistic values for a list of asset paths
        (
            ["resources/main.ts", "resources/styles.css"],
            None,
            '<script type="module" async="" defer="" src="/static/assets/main-l0sNRNKZ.js"></script><script type="module" async="" defer="" src="/static/assets/styles-l0sNRNKZ.js"></script>',
            "asset_multiple",
        ),
    ],
)
def test_get_asset_tag(
    vite_template_engine: ViteTemplateEngine,
    path: str | list[str],
    scripts_attrs: dict[str, Any] | None,
    expected_output: str,
    test_id: str,
) -> None:
    # Act
    result = str(vite_template_engine.get_asset_tag(path=path, scripts_attrs=scripts_attrs))

    # Assert
    assert result == expected_output


# Edge case tests for get_asset_tag
# Add more edge cases as needed
@pytest.mark.parametrize(
    "path, scripts_attrs, expected_output, test_id",
    [
        # Test with an empty path
        ("", None, "", "asset_empty_path"),
        # Test with empty scripts_attrs
        (
            "resources/main.ts",
            {},
            '<script type="module" async="" defer="" src="/static/assets/main-l0sNRNKZ.js"></script>',
            "asset_empty_attrs",
        ),
    ],
)
def test_get_asset_tag_edge_cases(
    vite_template_engine: ViteTemplateEngine,
    path: str | list[str],
    scripts_attrs: dict[str, Any] | None,
    expected_output: str,
    test_id: str,
) -> None:
    # Act
    if path == "resources/main.ts":
        result = vite_template_engine.get_asset_tag(path=path, scripts_attrs=scripts_attrs)

        # Assert
        assert result == expected_output, f"Failed test ID: {test_id}"
    else:
        with pytest.raises(KeyError) as exc_info:
            vite_template_engine.get_asset_tag(path, scripts_attrs)
        assert exc_info.type is KeyError


# Error case tests for get_asset_tag
# Add more error cases as needed
@pytest.mark.parametrize(
    "path, scripts_attrs, exception, test_id",
    [
        # Test with invalid path type
        (123, None, TypeError, "asset_invalid_path_type"),
        # Test with invalid scripts_attrs type
        ("resources/main.ts", "invalid", AttributeError, "asset_invalid_attrs_type"),
    ],
)
def test_get_asset_tag_error_cases(
    vite_template_engine: ViteTemplateEngine,
    path: str | list[str],
    scripts_attrs: dict[str, Any] | None,
    exception: Any,
    test_id: str,
) -> None:
    # Act / Assert
    with pytest.raises(exception) as exc_info:
        vite_template_engine.get_asset_tag(path, scripts_attrs)
    assert exc_info.type is exception



from __future__ import annotations

from typing import Any, Dict

import pytest
from litestar import Request, get
from litestar.middleware.session.server_side import ServerSideSessionConfig
from litestar.plugins.flash import (  # pyright: ignore[reportUnknownVariableType]  # pyright: ignore[reportUnknownVariableType]
    FlashConfig,
    FlashPlugin,
    flash,
)
from litestar.stores.memory import MemoryStore
from litestar.testing import create_test_client  # pyright: ignore[reportUnknownVariableType]

from litestar_vite.inertia import InertiaHeaders, InertiaPlugin
from litestar_vite.inertia.response import share
from litestar_vite.plugin import VitePlugin

pytestmark = pytest.mark.anyio


async def test_component_enabled(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/", component="Home")
    async def handler(request: Request[Any, Any, Any]) -> Dict[str, Any]:
        return {"thing": "value"}

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/")
        assert response.text.startswith("<!DOCTYPE html>")


async def test_component_inertia_header_enabled(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/", component="Home")
    async def handler(request: Request[Any, Any, Any]) -> Dict[str, Any]:
        return {"thing": "value"}

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/", headers={InertiaHeaders.ENABLED.value: "true"})
        assert (
            response.content
            == b'{"component":"Home","url":"/","version":"1.0","props":{"flash":{},"errors":{},"csrf_token":"","content":{"thing":"value"}}}'
        )


async def test_component_inertia_flash_header_enabled(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/", component="Home")
    async def handler(request: Request[Any, Any, Any]) -> Dict[str, Any]:
        flash(request, "a flash message", "info")
        return {"thing": "value"}

    with create_test_client(
        route_handlers=[handler],
        plugins=[
            inertia_plugin,
            vite_plugin,
            FlashPlugin(config=FlashConfig(template_config=vite_plugin.template_config)),
        ],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/", headers={InertiaHeaders.ENABLED.value: "true"})
        assert (
            response.content
            == b'{"component":"Home","url":"/","version":"1.0","props":{"flash":{"info":["a flash message"]},"errors":{},"csrf_token":"","content":{"thing":"value"}}}'
        )


async def test_component_inertia_shared_flash_header_enabled(
    inertia_plugin: InertiaPlugin,
    vite_plugin: VitePlugin,
) -> None:
    @get("/", component="Home")
    async def handler(request: Request[Any, Any, Any]) -> Dict[str, Any]:
        flash(request, "a flash message", "info")
        share(request, "auth", {"user": "nobody"})
        return {"thing": "value"}

    with create_test_client(
        route_handlers=[handler],
        plugins=[
            inertia_plugin,
            vite_plugin,
            FlashPlugin(config=FlashConfig(template_config=vite_plugin.template_config)),
        ],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/", headers={InertiaHeaders.ENABLED.value: "true"})
        assert (
            response.content
            == b'{"component":"Home","url":"/","version":"1.0","props":{"auth":{"user":"nobody"},"flash":{"info":["a flash message"]},"errors":{},"csrf_token":"","content":{"thing":"value"}}}'
        )


async def test_default_route_response_no_component(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/")
    async def handler(request: Request[Any, Any, Any]) -> Dict[str, Any]:
        return {"thing": "value"}

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get("/")
        assert response.content == b'{"thing":"value"}'


async def test_component_inertia_version_redirect(inertia_plugin: InertiaPlugin, vite_plugin: VitePlugin) -> None:
    @get("/", component="Home")
    async def handler(request: Request[Any, Any, Any]) -> Dict[str, Any]:
        return {"thing": "value"}

    with create_test_client(
        route_handlers=[handler],
        plugins=[inertia_plugin, vite_plugin],
        middleware=[ServerSideSessionConfig().middleware],
        stores={"sessions": MemoryStore()},
    ) as client:
        response = client.get(
            "/",
            headers={InertiaHeaders.ENABLED.value: "true", InertiaHeaders.VERSION.value: "wrong"},
        )
        assert (
            response.content
            == b'{"component":"Home","url":"/","version":"1.0","props":{"flash":{},"errors":{},"csrf_token":"","content":{"thing":"value"}}}'
        )



from pathlib import Path

from litestar_vite.config import ViteConfig


def test_default_vite_config() -> None:
    config = ViteConfig()
    assert isinstance(config.bundle_dir, Path)
    assert isinstance(config.public_dir, Path)
    assert config.ssr_output_dir is None
    assert isinstance(config.resource_dir, Path)
    assert config.root_dir is None



from litestar_vite.inertia.routes import EXCLUDED_METHODS, generate_js_routes
from tests.test_app.app import app


def test_route_export() -> None:
    routes = generate_js_routes(app=app)
    test_1 = {"POST", "OPTIONS"}
    test_2 = {"POST", "OPTIONS", "HEAD"}
    test_3 = {"HEAD"}
    test_4 = {"GET"}
    assert len(test_1.difference(EXCLUDED_METHODS)) == 1
    assert len(test_2.difference(EXCLUDED_METHODS)) == 1
    assert len(test_3.difference(EXCLUDED_METHODS)) == 0
    assert test_4.isdisjoint(EXCLUDED_METHODS)
    routes = generate_js_routes(app=app)
    json_response = routes.formatted_routes
    assert json_response
    assert routes is not None



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

__all__ = ("InertiaConfig",)


@dataclass
class InertiaConfig:
    """Configuration for InertiaJS support."""

    root_template: str = "index.html"
    """Name of the root template to use.

    This must be a path that is found by the Vite Plugin template config
    """
    component_opt_key: str = "component"
    """An identifier to use on routes to get the inertia component to render."""
    exclude_from_js_routes_key: str = "exclude_from_routes"
    """An identifier to use on routes to exclude a route from the generated routes typescript file."""
    redirect_unauthorized_to: str | None = None
    """Optionally supply a path where unauthorized requests should redirect."""
    extra_static_page_props: dict[str, Any] = field(default_factory=dict)
    """A dictionary of values to automatically add in to page props on every response."""
    extra_session_page_props: set[str] = field(default_factory=set)
    """A set of session keys for which the value automatically be added (if it exists) to the response."""



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, TypeVar

import markupsafe
from litestar.contrib.jinja import JinjaTemplateEngine
from litestar.template.base import TemplateEngineProtocol

from litestar_vite.loader import ViteAssetLoader

if TYPE_CHECKING:
    from pathlib import Path

    from jinja2 import Environment
    from jinja2 import Template as JinjaTemplate

    from litestar_vite.config import ViteConfig

T = TypeVar("T", bound=TemplateEngineProtocol["JinjaTemplate", Mapping[str, Any]])


class ViteTemplateEngine(JinjaTemplateEngine):
    """Jinja Template Engine with Vite Integration.

    This class extends :class:`litestar.contrib.jinja.JinjaTemplateEngine` to provide Vite asset integration
    and hot module reloading support.

    Raises:
        TemplateNotFoundException: If template is not found.
    """

    def __init__(
        self,
        directory: Path | list[Path] | None = None,
        engine_instance: Environment | None = None,
        config: ViteConfig | None = None,
    ) -> None:
        """Jinja2 based TemplateEngine.

        Args:
            directory: Direct path or list of directory paths from which to serve templates.
            engine_instance: A jinja Environment instance.
            config: Vite config
        """
        super().__init__(directory=directory, engine_instance=engine_instance)
        if config is None:
            msg = "Please configure the `ViteConfig` instance."
            raise ValueError(msg)
        self.config = config
        self.asset_loader = ViteAssetLoader.initialize_loader(config=self.config)
        self.engine.globals.update({"vite_hmr": self.get_hmr_client, "vite": self.get_asset_tag})  # pyright: ignore[reportCallIssue,reportUnknownMemberType,reportArgumentType]

    def get_hmr_client(self) -> markupsafe.Markup:
        """Generate the script tag for the Vite WS client for HMR.

        Only used when hot module reloading is enabled, in production this method returns an empty string.

        Returns:
            markupsafe.Markup: The script tag or an empty string.
        """
        return markupsafe.Markup(
            f"{self.asset_loader.generate_react_hmr_tags()}{self.asset_loader.generate_ws_client_tags()}",
        )

    def get_asset_tag(
        self,
        path: str | list[str],
        scripts_attrs: dict[str, str] | None = None,
        **_: Any,
    ) -> markupsafe.Markup:
        """Generate all assets include tags for the file in argument.

        Generates all scripts tags for this file and all its dependencies
        (JS and CSS) by reading the manifest file (for production only).
        In development Vite imports all dependencies by itself.
        Place this tag in <head> section of your page.

        Args:
            path: Path to a Vite asset to include.
            scripts_attrs: Dictionary of attributes to add to script tags.
            **_: Additional keyword arguments (ignored).

        Returns:
            markupsafe.Markup: HTML markup containing all required asset tags.
        """
        if isinstance(path, str):
            path = [path]
        return markupsafe.Markup(
            "".join([self.asset_loader.generate_asset_tags(p, scripts_attrs=scripts_attrs) for p in path]),
        )

    @classmethod
    def from_environment(cls, config: ViteConfig, jinja_environment: Environment) -> ViteTemplateEngine:  # type: ignore[override]
        """Create a JinjaTemplateEngine from an existing jinja Environment instance.

        Args:
            config: Vite config
            jinja_environment (jinja2.environment.Environment): A jinja Environment instance.

        Returns:
            JinjaTemplateEngine instance
        """
        return cls(directory=None, config=config, engine_instance=jinja_environment)



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING
from urllib.parse import unquote

from litestar import Request
from litestar.connection.base import (
    AuthT,
    StateT,
    UserT,
    empty_receive,
    empty_send,
)

from litestar_vite.inertia._utils import InertiaHeaders

__all__ = ("InertiaDetails", "InertiaRequest")


if TYPE_CHECKING:
    from litestar.types import Receive, Scope, Send


class InertiaDetails:
    """InertiaDetails holds all the values sent by Inertia client in headers and provide convenient properties."""

    def __init__(self, request: Request[UserT, AuthT, StateT]) -> None:
        """Initialize :class:`InertiaDetails`"""
        self.request = request

    def _get_header_value(self, name: InertiaHeaders) -> str | None:
        """Parse request header

        Check for uri encoded header and unquotes it in readable format.
        """

        if value := self.request.headers.get(name.value.lower()):
            is_uri_encoded = self.request.headers.get(f"{name.value.lower()}-uri-autoencoded") == "true"
            return unquote(value) if is_uri_encoded else value
        return None

    def _get_route_component(self) -> str | None:
        """Get the route component.

        Checks for the `component` key within the route  configuration.
        """
        rh = self.request.scope.get("route_handler")  # pyright: ignore[reportUnknownMemberType]
        if rh:
            return rh.opt.get("component")
        return None

    def __bool__(self) -> bool:
        """Check if request is sent by an Inertia client."""
        return self._get_header_value(InertiaHeaders.ENABLED) == "true"

    @cached_property
    def route_component(self) -> str | None:
        """Partial Data Reload."""
        return self._get_route_component()

    @cached_property
    def partial_component(self) -> str | None:
        """Partial Data Reload."""
        return self._get_header_value(InertiaHeaders.PARTIAL_COMPONENT)

    @cached_property
    def partial_data(self) -> str | None:
        """Partial Data Reload."""
        return self._get_header_value(InertiaHeaders.PARTIAL_DATA)

    @cached_property
    def referer(self) -> str | None:
        """Partial Data Reload."""
        return self._get_header_value(InertiaHeaders.REFERER)

    @cached_property
    def is_partial_render(self) -> bool:
        """Is Partial Data Reload."""
        return bool(self.partial_component == self.route_component and self.partial_data)

    @cached_property
    def partial_keys(self) -> list[str]:
        """Is Partial Data Reload."""
        return self.partial_data.split(",") if self.partial_data is not None else []


class InertiaRequest(Request[UserT, AuthT, StateT]):
    """Inertia Request class to work with Inertia client."""

    __slots__ = ("inertia",)

    def __init__(self, scope: Scope, receive: Receive = empty_receive, send: Send = empty_send) -> None:
        """Initialize :class:`InertiaRequest`"""
        super().__init__(scope=scope, receive=receive, send=send)
        self.inertia = InertiaDetails(self)

    @property
    def is_inertia(self) -> bool:
        """True if the request contained inertia headers."""
        return bool(self.inertia)

    @property
    def inertia_enabled(self) -> bool:
        """True if the route handler contains an inertia enabled configuration."""
        return bool(self.inertia.route_component is not None)

    @property
    def is_partial_render(self) -> bool:
        """True if the route handler contains an inertia enabled configuration."""
        return self.inertia.is_partial_render

    @property
    def partial_keys(self) -> set[str]:
        """True if the route handler contains an inertia enabled configuration."""
        return set(self.inertia.partial_keys)
